// ST60 LOG SPEED SIMULATOR: NMEA functions
// Copyright December 2015 by Nigel Mercier www.mercier.org.uk

#include "NMEAfn.h"
// #include "NMEATest.h"        // DEGUG/TEST

// Global constants
const char EOL = '\r' ;              // \r is CR escape sequence for end of NMEA sentence

// Global variables for NMEA interface
volatile char speed[12] ;           // quick storage for speed data
volatile unsigned int time_Out ;    // set to 65535 before call to parseNMEA()

// Global variables not needed for NMEA interface
volatile char field[12] ;           // temporary buffer for each NMEA field
volatile unsigned short errorF ;    // flag > 0 if RX string runout error
volatile unsigned short checksum ;  // running checksum from NMEA input

// NMEA functions (see NMEAfn.h for list)
void initNMEA(void)
{ // initialise UART
  UART1_Init(4800);  // Baud = 4800
  Delay_ms(100) ;    // allow to stabilise
}

unsigned short waitChar(void)
{ // wait for UART to be ready, return next character from NMEA string
  //return waitCharTest() ;  // DEBUG & TESTING, comment out remaining lines, see NMEATest.c
  unsigned short t = 0 ;
  while(!UART1_Data_Ready())   // wait here until UART data ready, or time_Out runout
  {
    t++ ;                      // increment outer count
    if (t == 25)               // 25 gives 10 seconds if timeOut set to 65535 in main()
    {
      time_Out-- ;             // decrement inner count (no ROM room for long int)
      t = 0 ;                  // reset outer count
    }
    if (!time_Out)             // if no NMEA data ready after long delay, time_Out gets to 0
      return 0xFF ;            // 0xFF is secondary error flag if needed
  }
  return UART1_Read() ;        // return character from UART
}

void getField(char delimiter, char max)
{ // call waitChar() until delimiter, or runout (characters > max)
  unsigned short i, c , field_End ;
  i = 0 ;               // index into field string
  field_End = 1 ;       // becomes 0 when delimiter found (in range of max)
  
// Read until delimiter is detected, or max number of characters for field
  while((field_End)&&(i <= max)) // = to allow 1 char for terminator
  { // fieldEnd is 1 if delimiter found, max = max string length + terminator
    c = waitChar() ;   // wait for next character from UART
    if (!time_Out)     // timeout error?
      break ;          // return with time_Out at zero
    if(!((delimiter == EOL)||(c =='*'))) // delimiter not EOL AND char not '*'
    { // update checksum except for last 2 digits and '*' delimiter
      checksum ^= c ;                    // shorthand for checksum = checksum^c ;
    }
    if(c == delimiter)  // end of field delimiter?
    {
      field_End = 0;    // found delimiter, flag 0 ready for exit
      c = 0 ;           // ready to store \0 (zero terminator) instead of delimiter
    }
    field[i++] = c ;    // store new character in field[] string, increment index
  }
  errorF += field_End ; // flag runout error if field_End not zero
}

void waitHdr(void)
{ 
  char gpsHdr[]="$GPRMC," ;  // string to search for in UART RX stream
  unsigned short flag, i ;   // f to flag match, i = loop variable
  char c ;                   // c is character from UART
  i = 0 ;                    // first character in wanted header
  flag = 0 ;                 // becomes 1 when full string matched
  
  // read UART forever until header "$GPRMC," is detected
  while(!flag)
  {
    c = waitChar() ;   // wait for character from UART
    if (!time_Out)
      break ;          // return with time_Out at zero
    if(c == gpsHdr[i]) // does character match RMC header?
    {
      i++;             // character match, get ready for next character
      if(i > 6)        // 6 characters to match
        flag = 1;      // past end so all matched, so prepare to exit
    }
    else i = 0;        // character mismatch, flag still 0, start while loop again
  }
}

unsigned short HexDigit(char c)
{ // return ASCII hex nibble to number for checksum
  c.F5 = 1;                       // set bit 5, convert any A...F to lowercase
  if (c >= '0' && c <= '9')       // test for numeric
    return c - '0' ;              // return as number 0...9
  else if (c >= 'a' && c <= 'f')  // test for a...f
    return 10 + c - 'a' ;         // return as number 10...15
  else
    return 0 ;                    // not valid
}

unsigned short getHexNum(void)
{ // return field[0...1] as number 0...255
  return ((HexDigit(field[0]) << 4) | HexDigit(field[1])) ;
}

unsigned short parseNMEA(void)
{ // parse 1 sentence, extract data as required, see SOG (speedE2) for example
  // GPS Status flag & checksum processed, fields ONLY checked for over-length

  char statusF ;          // Status byte, 'A' is OK
  unsigned short check2 ; // zero if calculated checksum == last two characters
  
  waitHdr() ;             // wait for NMEA header "$GPRMC,"
  if (!time_Out)
    return 0xFF ;         // return with 0xFF if time_Out at zero
  checksum = 103 ;        // preload checksum: XOR of "GPRMC," as already received
  // NMEA header "$GPRMC," loaded, ready to parse remaining fields into field[12] variable
  errorF = 0 ;            // error flag > 0 if RX string runout (too many chars)
  // following NMEA fields terminated with ',' (comma) with defined max length
  getField(',' , 10) ;    // UTC Time (float hhmmss.sss = 10 char)
  getField(',' , 1) ;     // Status (1 char): 'A' = valid, 'V' = invalid (checked!)
  // copy status flag to statusF
  statusF = field[0] ;    // status flag saved as char for error return
  getField(',', 10) ;     // Latitude (float dddmm.mmmm = 10 char)
  getField(',', 1) ;      // 'N' or 'S' (1 char)
  getField(',', 10) ;     // Longitude (float dddmm.mmmm = 10 char)
  getField(',', 1) ;      // 'E' or 'W' (1 char)
  getField(',', 9) ;      // SOG knots (float kkkkk.kkk = 9 char)
  copySpeed(field) ;      // copy field[12] to speed[] up to '\0' terminator

  getField(',', 6) ;      // COG True (float ddd.dd = 6 char)
  getField(',', 6) ;      // Date (ddmmyy) (6 char)
  getField(',', 5) ;      // Magnetic Var (dd.dd = 5 char)
  // following NMEA fields terminated with '*' or EOL (<CR> or <CR><LF>) within max
  getField('*', 1) ;      // Magnetic Var 'E' or 'W' (1 char) terminates with '*'
  getField(EOL, 2) ;      // Checksum (2 hex char) terminated with EOL
  check2 = (getHexNum() != checksum) ; // getHexNum returns field[0...1] as #
  return ( (statusF != 'A') + (check2) + (errorF) ) ; //  return 0 if no errors
}

void copySpeed(char *f)
{  // copy field[12] to speed[] up to '\0' terminator
  unsigned short i = 0 ;      // loop variable used in string copy
 // i = 0 ;
  while(f[i] != '\0')
    {  speed[i] = f[i] ;
       i++ ;
    }  speed[i] = '\0' ;  // store \0 (zero terminator) to end speed[] string
}

unsigned int strToInt(char *s)
{ // returns integer value of argument, times 100 to lose DP
  float result = 0.0 ;
  float scale  = 0.0 ;  // if DP found, counts divisor
  while(*s) // do while not \0 string terminator
  {
    if((*s >= '0') && (*s <= '9')) // if numeric character ...
    { // ... shift existing result to next decimal column, add current character cast as float
      result = 10.0 * result + (float)(*s - '0') ;
      scale *= 10.0 ;      // scale only changes when scale NOT 0.0, so after DP found
    }

    else if(*s == '.') // not numeric, so if DP ...
      scale = 1.0 ;    // ... start building divisor for use at exit

    else // not numeric or DP so error
      return 0 ; // error in field, so flag zero speed

    s++ ; // next character

  } // exit at end of string \0
    // divide by 1.0 if scale is 0.0, or by scale if scale not 0.0 due to DP found
    result = result / (scale == 0.0 ? 1.0 : scale) ;
    return (int)(100*(result+0.005)) ; // round 3rd decimal (if any) return integer
}